@@ -786,7 +786,7 @@ static int verify_krb_v4_tgt (realm)
     struct hostent *hp;
     KTEXT_ST ticket;
     AUTH_DAT authdata;
-    unsigned long addr;
+    unsigned KRB4_32 addr;
     static /*const*/ char rcmd_str[] = "rcmd";
 #if 0
     char key[8];
@@ -1861,7 +1861,7 @@ int main(argc, argv)
     (void) strncpy(tbuf+1, p?(p+1):pwd->pw_shell, sizeof(tbuf) - 1);
     tbuf[sizeof(tbuf) - 1] = '\0';
 
-    execlp(pwd->pw_shell, tbuf, 0);
+    execlp(pwd->pw_shell, tbuf, (char *)NULL);
     fprintf(stderr, "login: no shell: ");
     perror(pwd->pw_shell);
     exit(0);
