@@ -79,7 +79,7 @@
 #include <assert.h>
 
 /*
- * $Id: init_sec_context.c,v 1.84.4.1 2005/01/13 21:55:10 tlyu Exp $
+ * $Id: init_sec_context.c 18131 2006-06-14 22:27:54Z tlyu $
  */
 
 /* XXX This is for debugging only!!!  Should become a real bitfield
@@ -155,7 +155,7 @@ make_gss_checksum (krb5_context context, krb5_auth_context auth_context,
     unsigned char *ptr;
     struct gss_checksum_data *data = cksum_data;
     krb5_data credmsg;
-    int junk;
+    unsigned int junk;
 
     data->checksum_data.data = 0;
     credmsg.data = 0;
@@ -325,7 +325,7 @@ make_ap_req_v1(context, ctx, cred, k_cred, chan_bindings, mech_type, token)
 
    ptr = t;
 
-   g_make_token_header((gss_OID) mech_type, ap_req.length,
+   g_make_token_header(mech_type, ap_req.length,
 		       &ptr, KG_TOK_CTX_AP_REQ);
 
    TWRITE_STR(ptr, (unsigned char *) ap_req.data, ap_req.length);
@@ -720,7 +720,7 @@ mutual_auth(
 
    ptr = (unsigned char *) input_token->value;
 
-   if (g_verify_token_header((gss_OID) ctx->mech_used,
+   if (g_verify_token_header(ctx->mech_used,
 			     &(ap_rep.length),
 			     &ptr, KG_TOK_CTX_AP_REP,
 			     input_token->length, 1)) {
@@ -847,7 +847,7 @@ krb5_gss_init_sec_context(minor_status, claimant_cred_handle,
    OM_uint32 tmp_min_stat;
 
    if (*context_handle == GSS_C_NO_CONTEXT) {
-       kerr = krb5_init_context(&context);
+       kerr = krb5_gss_init_context(&context);
        if (kerr) {
 	   *minor_status = kerr;
 	   return GSS_S_FAILURE;
@@ -918,6 +918,9 @@ krb5_gss_init_sec_context(minor_status, claimant_cred_handle,
    } else if (g_OID_equal(mech_type, gss_mech_krb5_old)) {
        if (!cred->prerfc_mech)
 	   err = 1;
+   } else if (g_OID_equal(mech_type, gss_mech_krb5_wrong)) {
+       if (!cred->rfc_mech)
+	   err = 1;
    } else {
        err = 1;
    }
@@ -966,3 +969,42 @@ krb5_gss_init_sec_context(minor_status, claimant_cred_handle,
 
    return(major_status);
 }
+
+k5_mutex_t kg_kdc_flag_mutex = K5_MUTEX_PARTIAL_INITIALIZER;
+static int kdc_flag = 0;
+
+krb5_error_code
+krb5_gss_init_context (krb5_context *ctxp)
+{
+    krb5_error_code err;
+    int is_kdc;
+
+    err = gssint_initialize_library();
+    if (err)
+	return err;
+    err = k5_mutex_lock(&kg_kdc_flag_mutex);
+    if (err)
+	return err;
+    is_kdc = kdc_flag;
+    k5_mutex_unlock(&kg_kdc_flag_mutex);
+    if (is_kdc)
+	return krb5int_init_context_kdc(ctxp);
+    else
+	return krb5_init_context(ctxp);
+}
+
+krb5_error_code
+krb5_gss_use_kdc_context()
+{
+    krb5_error_code err;
+
+    err = gssint_initialize_library();
+    if (err)
+	return err;
+    err = k5_mutex_lock(&kg_kdc_flag_mutex);
+    if (err)
+	return err;
+    kdc_flag = 1;
+    k5_mutex_unlock(&kg_kdc_flag_mutex);
+    return 0;
+}
