@@ -1,7 +1,7 @@
 /*
  * patchlevel.h
  *
- * Copyright (C) 2004 by the Massachusetts Institute of Technology.
+ * Copyright (C) 2004-2006 by the Massachusetts Institute of Technology.
  * All rights reserved.
  *
  * Export of this software from the United States of America may
@@ -37,7 +37,7 @@
  * will be incremented, and KRB5_RELTAIL will revert to "prerelease".
  *
  * KRB5_RELTAG contains the CVS tag name corresponding to the release.
-#define KRB5_RELDATE "20051116"
+#define KRB5_RELDATE "20060701"
  * normally be undefined for checked-in code.
  */
 
@@ -51,8 +51,8 @@
  * organization.
  */
 #define KRB5_MAJOR_RELEASE 1
-#define KRB5_MINOR_RELEASE 4
-#define KRB5_PATCHLEVEL 3
+#define KRB5_MINOR_RELEASE 5
+#define KRB5_PATCHLEVEL 0
 /* #undef KRB5_RELTAIL */
-#define KRB5_RELDATE "20051116"
-#define KRB5_RELTAG "krb5-1-4-3-final"
+#define KRB5_RELDATE "20060701"
+#define KRB5_RELTAG "tags/krb5-1-5-final"
