@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2004 Massachusetts Institute of Technology
+ * Copyright (c) 2005 Massachusetts Institute of Technology
  *
  * Permission is hereby granted, free of charge, to any person
  * obtaining a copy of this software and associated documentation
@@ -25,6 +25,9 @@
 /* $Id$ */
 
 #include<kmminternal.h>
+#ifdef DEBUG
+#include<assert.h>
+#endif
 
 static LONG pending_modules = 0;
 static LONG pending_plugins = 0;
@@ -38,6 +41,9 @@ kmmint_check_completion(void) {
         InterlockedIncrement(&startup_signal) == 1) {
 
         load_done = TRUE;
+
+        /* TODO: check for orphaned plugins */
+
         kmq_post_message(KMSG_KMM, KMSG_KMM_I_DONE, 0, 0);
     }
 }
@@ -56,14 +62,25 @@ kmmint_remove_from_module_queue(void) {
 }
 
 void
-kmmint_add_to_plugin_queue(void) {
-    InterlockedIncrement(&pending_plugins);
+kmmint_add_to_plugin_queue(kmm_plugin_i * plugin) {
+    EnterCriticalSection(&cs_kmm);
+    if (!(plugin->flags & KMM_PLUGIN_FLAG_IN_QUEUE)) {
+        InterlockedIncrement(&pending_plugins);
+        plugin->flags |= KMM_PLUGIN_FLAG_IN_QUEUE;
+    }
+    LeaveCriticalSection(&cs_kmm);
 }
 
 void
-kmmint_remove_from_plugin_queue(void) {
-    InterlockedDecrement(&pending_plugins);
+kmmint_remove_from_plugin_queue(kmm_plugin_i * plugin) {
+    EnterCriticalSection(&cs_kmm);
 
+    if (plugin->flags & KMM_PLUGIN_FLAG_IN_QUEUE) {
+        InterlockedDecrement(&pending_plugins);
+        plugin->flags &= ~KMM_PLUGIN_FLAG_IN_QUEUE;
+    }
+
+    LeaveCriticalSection(&cs_kmm);
     kmmint_check_completion();
 }
 
@@ -74,11 +91,10 @@ kmm_load_pending(void) {
 
 /*! \internal
   \brief Message handler for the registrar thread. */
-khm_boolean KHMAPI kmm_reg_cb(
-    khm_int32 msg_type, 
-    khm_int32 msg_sub_type, 
-    khm_ui_4 uparam,
-    void *vparam)
+khm_boolean KHMAPI kmmint_reg_cb(khm_int32 msg_type, 
+                                 khm_int32 msg_sub_type, 
+                                 khm_ui_4 uparam,
+                                 void *vparam)
 {
     /* we should only be getting <KMSG_KMM,KMSG_KMM_I_REG> anyway */
     if(msg_type != KMSG_KMM || msg_sub_type != KMSG_KMM_I_REG)
@@ -86,22 +102,22 @@ khm_boolean KHMAPI kmm_reg_cb(
 
     switch(uparam) {
         case KMM_REG_INIT_MODULE:
-            kmm_init_module((kmm_module_i *) vparam);
+            kmmint_init_module((kmm_module_i *) vparam);
             kmm_release_module(kmm_handle_from_module((kmm_module_i *) vparam));
             break;
 
         case KMM_REG_EXIT_MODULE:
-            kmm_exit_module((kmm_module_i *) vparam);
+            kmmint_exit_module((kmm_module_i *) vparam);
             kmm_release_module(kmm_handle_from_module((kmm_module_i *) vparam));
             break;
 
         case KMM_REG_INIT_PLUGIN:
-            kmm_init_plugin((kmm_plugin_i *) vparam);
+            kmmint_init_plugin((kmm_plugin_i *) vparam);
             kmm_release_plugin(kmm_handle_from_plugin((kmm_plugin_i *) vparam));
             break;
 
         case KMM_REG_EXIT_PLUGIN:
-            kmm_exit_plugin((kmm_plugin_i *) vparam);
+            kmmint_exit_plugin((kmm_plugin_i *) vparam);
             kmm_release_plugin(kmm_handle_from_plugin((kmm_plugin_i *) vparam));
             break;
     }
@@ -112,15 +128,13 @@ khm_boolean KHMAPI kmm_reg_cb(
   \brief The registrar thread.
 
   The only thing this function does is to dispatch messages to the
-  callback routine ( kmm_reg_cb() ) */
-DWORD WINAPI kmm_registrar(
-  LPVOID lpParameter
-)
+  callback routine ( kmmint_reg_cb() ) */
+DWORD WINAPI kmmint_registrar(LPVOID lpParameter)
 {
     tid_registrar = GetCurrentThreadId();
 
-    kmq_subscribe(KMSG_KMM, kmm_reg_cb);
-    kmq_subscribe(KMSG_SYSTEM, kmm_reg_cb);
+    kmq_subscribe(KMSG_KMM, kmmint_reg_cb);
+    kmq_subscribe(KMSG_SYSTEM, kmmint_reg_cb);
 
     SetEvent(evt_startup);
 
@@ -137,47 +151,89 @@ DWORD WINAPI kmm_registrar(
   Each plugin gets its own plugin thread which is used to dispatch
   messages to the plugin.  This acts as the thread function for the
   plugin thread.*/
-DWORD WINAPI kmm_plugin_broker(LPVOID lpParameter)
+DWORD WINAPI kmmint_plugin_broker(LPVOID lpParameter)
 {
     DWORD rv = 0;
     kmm_plugin_i * p = (kmm_plugin_i *) lpParameter;
 
+    _begin_task(0);
+    _report_mr1(KHERR_NONE, MSG_PB_START, _cstr(p->p.name));
+    _describe();
+
     TlsSetValue(tls_kmm, (LPVOID) p);
 
     kmm_hold_plugin(kmm_handle_from_plugin(p));
 
     p->tid_thread = GetCurrentThreadId();
 
-    rv = (p->p.msg_proc(KMSG_SYSTEM, KMSG_SYSTEM_INIT, 0, (void *) &(p->p)));
+    if (IsBadCodePtr(p->p.msg_proc)) {
+        _report_mr0(KHERR_WARNING, MSG_PB_INVALID_CODE_PTR);
+        rv = KHM_ERROR_INVALID_PARAM;
+    } else {
+        rv = (*p->p.msg_proc)(KMSG_SYSTEM, KMSG_SYSTEM_INIT, 
+                              0, (void *) &(p->p));
+        _report_mr1(KHERR_INFO, MSG_PB_INIT_RV, _int32(rv));
+    }
 
     /* if it fails to initialize, we exit the plugin */
     if(KHM_FAILED(rv)) {
-	kmmint_remove_from_plugin_queue();
+
+        kherr_report(KHERR_ERROR,
+                     (wchar_t *) MSG_PB_INIT_FAIL_S,
+                     (wchar_t *) KHERR_FACILITY,
+                     NULL,
+                     (wchar_t *) MSG_PB_INIT_FAIL,
+                     (wchar_t *) MSG_PB_INIT_FAIL_G,
+                     KHERR_FACILITY_ID,
+                     KHERR_SUGGEST_NONE,
+                     _cstr(p->p.name),
+                     _cstr(p->p.description),
+                     _cstr(p->module->path),
+                     _cstr(p->module->support),
+                     KHERR_RF_MSG_SHORT_DESC |
+                     KHERR_RF_MSG_LONG_DESC |
+                     KHERR_RF_MSG_SUGGEST
+#ifdef _WIN32
+                     ,KHERR_HMODULE
+#endif
+                     );
+        _resolve();
+
+        /* exit the plugin first.  Otherwise it may not uninitialize correctly */
+        (*p->p.msg_proc)(KMSG_SYSTEM, KMSG_SYSTEM_EXIT, 0, (void *) &(p->p));
+
+        kmmint_remove_from_plugin_queue(p);
         rv = 1;
+        _end_task();
+
+        p->state = KMM_PLUGIN_STATE_FAIL_INIT;
         goto _exit;
     }
 
     /* subscribe to default message classes by plugin type */
-    if(p->p.type & KHM_PITYPE_CRED) {
+    if(p->p.type == KHM_PITYPE_CRED) {
         kmq_subscribe(KMSG_SYSTEM, p->p.msg_proc);
         kmq_subscribe(KMSG_KCDB, p->p.msg_proc);
         kmq_subscribe(KMSG_CRED, p->p.msg_proc);
-    }
-
-    if(p->p.flags & KHM_PIFLAG_IDENTITY_PROVIDER) {
+    } else if(p->p.type == KHM_PITYPE_IDENT) {
         khm_handle h = NULL;
 
+        kmq_subscribe(KMSG_SYSTEM, p->p.msg_proc);
+        kmq_subscribe(KMSG_KCDB, p->p.msg_proc);
+
         kmq_create_subscription(p->p.msg_proc, &h);
         kcdb_identity_set_provider(h);
         /* kcdb deletes the subscription when it's done with it */
-    }
-
-    if(p->p.type == KHM_PITYPE_CONFIG) {
+    } else if(p->p.type == KHM_PITYPE_CONFIG) {
         /*TODO: subscribe to configuration provider messages here */
     }
 
     p->state = KMM_PLUGIN_STATE_RUNNING;
 
+    _report_mr0(KHERR_INFO, MSG_PB_INIT_DONE);
+
+    _end_task();
+
     /* if there were any plugins that were waiting for this one to
        start, we should start them too */
     EnterCriticalSection(&cs_kmm);
@@ -191,6 +247,7 @@ DWORD WINAPI kmm_plugin_broker(LPVOID lpParameter)
             pd->n_unresolved--;
 
             if(pd->n_unresolved == 0) {
+                kmmint_add_to_plugin_queue(pd);
                 kmm_hold_plugin(kmm_handle_from_plugin(pd));
                 kmq_post_message(KMSG_KMM, KMSG_KMM_I_REG, KMM_REG_INIT_PLUGIN, (void *) pd);
             }
@@ -198,33 +255,33 @@ DWORD WINAPI kmm_plugin_broker(LPVOID lpParameter)
     } while(FALSE);
     LeaveCriticalSection(&cs_kmm);
 
-    kmmint_remove_from_plugin_queue();
+    kmmint_remove_from_plugin_queue(p);
 
     /* main message loop */
     while(KHM_SUCCEEDED(kmq_dispatch(INFINITE)));
 
     /* unsubscribe from default message classes by plugin type */
-    if(p->p.type & KHM_PITYPE_CRED) {
+    if(p->p.type == KHM_PITYPE_CRED) {
         kmq_unsubscribe(KMSG_SYSTEM, p->p.msg_proc);
         kmq_unsubscribe(KMSG_KCDB, p->p.msg_proc);
         kmq_unsubscribe(KMSG_CRED, p->p.msg_proc);
-    }
-
-    if(p->p.flags & KHM_PIFLAG_IDENTITY_PROVIDER) {
+    } else if (p->p.type == KHM_PITYPE_IDENT) {
+        kmq_unsubscribe(KMSG_KCDB, p->p.msg_proc);
+        kmq_unsubscribe(KMSG_SYSTEM, p->p.msg_proc);
         kcdb_identity_set_provider(NULL);
-    }
-
-    if(p->p.type == KHM_PITYPE_CONFIG) {
+    } else if(p->p.type == KHM_PITYPE_CONFIG) {
         /*TODO: unsubscribe from configuration provider messages here */
     }
 
     p->p.msg_proc(KMSG_SYSTEM, KMSG_SYSTEM_EXIT, 0, (void *) &(p->p));
 
-_exit:
-    p->state = KMM_PLUGIN_STATE_EXITED;
+ _exit:
+    if (p->state >= 0)
+        p->state = KMM_PLUGIN_STATE_EXITED;
 
     /* the following call will automatically release the plugin */
-    kmq_post_message(KMSG_KMM, KMSG_KMM_I_REG, KMM_REG_EXIT_PLUGIN, (void *) p);
+    kmq_post_message(KMSG_KMM, KMSG_KMM_I_REG, 
+                     KMM_REG_EXIT_PLUGIN, (void *) p);
 
     TlsSetValue(tls_kmm, (LPVOID) 0);
 
@@ -237,17 +294,17 @@ _exit:
 /*! \internal
   \brief Initialize a plugin
 
-  \note If kmm_init_plugin() is called on a plugin, then kmm_exit_plugin()
+  \note If kmmint_init_plugin() is called on a plugin, then kmmint_exit_plugin()
       \b must be called for the plugin.
 
   \note Should only be called from the context of the registrar thread */
-void kmm_init_plugin(kmm_plugin_i * p) {
+void kmmint_init_plugin(kmm_plugin_i * p) {
     DWORD dummy;
     khm_handle csp_plugin   = NULL;
     khm_handle csp_plugins  = NULL;
     khm_int32 t;
 
-    /* the following will be undone in kmm_exit_plugin() */
+    /* the following will be undone in kmmint_exit_plugin() */
     kmm_hold_plugin(kmm_handle_from_plugin(p));
 
     EnterCriticalSection(&cs_kmm);
@@ -272,6 +329,10 @@ void kmm_init_plugin(kmm_plugin_i * p) {
     }
 
     p->state = KMM_PLUGIN_STATE_PREINIT;
+
+    kmmint_delist_plugin(p);
+    kmmint_list_plugin(p);
+
     LeaveCriticalSection(&cs_kmm);
 
     if(KHM_FAILED(kmm_get_plugins_config(0, &csp_plugins))) {
@@ -282,8 +343,7 @@ void kmm_init_plugin(kmm_plugin_i * p) {
     }
 
     if(KHM_FAILED(kmm_get_plugin_config(p->p.name, 0, &csp_plugin)) ||
-        KHM_FAILED(khc_read_int32(csp_plugin, L"Flags", &t)))
-    {
+       KHM_FAILED(khc_read_int32(csp_plugin, L"Flags", &t))) {
         if(KHM_FAILED(kmm_register_plugin(&(p->p), 0))) {
             _report_mr0(KHERR_ERROR, MSG_IP_NOT_REGISTERED);
 
@@ -304,11 +364,11 @@ void kmm_init_plugin(kmm_plugin_i * p) {
             p->state = KMM_PLUGIN_STATE_FAIL_NOT_REGISTERED;
             goto _exit;
         }
+
     }
 
     if(t & KMM_PLUGIN_FLAG_DISABLED) {
-        _report_mr0(KHERR_ERROR, MSG_IP_DISABLED);
-
+        p->flags |= KMM_PLUGIN_FLAG_DISABLED;
         p->state = KMM_PLUGIN_STATE_FAIL_DISABLED;
         goto _exit;
     }
@@ -329,13 +389,15 @@ void kmm_init_plugin(kmm_plugin_i * p) {
         wchar_t * d;
         khm_size sz = 0;
 
-        if(khc_read_multi_string(csp_plugin, L"Dependencies", NULL, &sz) != KHM_ERROR_TOO_LONG)
+        if(khc_read_multi_string(csp_plugin, L"Dependencies", 
+                                 NULL, &sz) != KHM_ERROR_TOO_LONG)
             break;
 
-        deps = malloc(sz);
-        if(KHM_FAILED(khc_read_multi_string(csp_plugin, L"Dependencies", deps, &sz))) {
+        deps = PMALLOC(sz);
+        if(KHM_FAILED(khc_read_multi_string(csp_plugin, L"Dependencies", 
+                                            deps, &sz))) {
             if(deps)
-                free(deps);
+                PFREE(deps);
             break;
         }
 
@@ -343,14 +405,14 @@ void kmm_init_plugin(kmm_plugin_i * p) {
             kmm_plugin_i * pd;
             int i;
 
-            pd = kmm_get_plugin_i(d);
+            pd = kmmint_get_plugin_i(d);
 
             if(pd->state == KMM_PLUGIN_STATE_NONE) {
                 /* the dependant was not previously known */
                 pd->state = KMM_PLUGIN_STATE_PLACEHOLDER;
             }
 
-            for(i=0; i<pd->n_dependants; i++) {
+            for(i=0; i < pd->n_dependants; i++) {
                 if(pd->dependants[i] == p)
                     break;
             }
@@ -361,7 +423,7 @@ void kmm_init_plugin(kmm_plugin_i * p) {
                     RaiseException(1, EXCEPTION_NONCONTINUABLE, 0, NULL);
                 }
 
-                /* released in kmm_free_plugin() */
+                /* released in kmmint_free_plugin() */
                 kmm_hold_plugin(kmm_handle_from_plugin(p));
                 pd->dependants[pd->n_dependants] = p;
                 pd->n_dependants++;
@@ -378,15 +440,12 @@ void kmm_init_plugin(kmm_plugin_i * p) {
             p->state = KMM_PLUGIN_STATE_HOLD;
         }
 
-        free(deps);
+        PFREE(deps);
 
     } while(FALSE);
-    LeaveCriticalSection(&cs_kmm);
 
-    EnterCriticalSection(&cs_kmm);
     p->module->plugin_count++;
-    kmm_delist_plugin(p);
-    kmm_list_plugin(p);
+
     LeaveCriticalSection(&cs_kmm);
 
     if(p->state == KMM_PLUGIN_STATE_HOLD) {
@@ -395,15 +454,14 @@ void kmm_init_plugin(kmm_plugin_i * p) {
         goto _exit_post;
     }
 
-    kmmint_add_to_plugin_queue();
+    kmmint_add_to_plugin_queue(p);
 
-    p->ht_thread = CreateThread(
-        NULL,
-        0,
-        kmm_plugin_broker,
-        (LPVOID) p,
-        CREATE_SUSPENDED,
-        &dummy);
+    p->ht_thread = CreateThread(NULL,
+                                0,
+                                kmmint_plugin_broker,
+                                (LPVOID) p,
+                                CREATE_SUSPENDED,
+                                &dummy);
 
     p->state = KMM_PLUGIN_STATE_INIT;
 
@@ -448,18 +506,20 @@ _exit:
   linked list and hashtable, it also frees up p.
    
   \note Should only be called from the context of the registrar thread. */
-void kmm_exit_plugin(kmm_plugin_i * p) {
+void kmmint_exit_plugin(kmm_plugin_i * p) {
     int np;
+    khm_boolean release_plugin = TRUE;
 
     if(p->state == KMM_PLUGIN_STATE_RUNNING ||
-        p->state == KMM_PLUGIN_STATE_INIT)
-    {
+       p->state == KMM_PLUGIN_STATE_INIT) {
+
         kmq_post_thread_quit_message(p->tid_thread, 0, NULL);
         /* when we post the quit message to the plugin thread, the plugin
            broker terminates the plugin and posts a EXIT_PLUGIN message,
            which calls this function again.  We just exit here because
            the EXIT_PLUGIN message will end up calling us again momentarily */
         return;
+
     }
 
     if(p->ht_thread) {
@@ -470,16 +530,25 @@ void kmm_exit_plugin(kmm_plugin_i * p) {
 
     EnterCriticalSection(&cs_kmm);
 
-    /* undo reference count done in kmm_init_plugin() */
+    /* undo reference count done in kmmint_init_plugin() */
     if(p->state == KMM_PLUGIN_STATE_EXITED ||
-        p->state == KMM_PLUGIN_STATE_HOLD) 
-    {
+       p->state == KMM_PLUGIN_STATE_HOLD) {
+
         np = --(p->module->plugin_count);
+
     } else {
         /* the plugin was never active.  We can't base a module unload
            decision on np */
         np = TRUE;
     }
+
+    /* The plugin is in an error state.  We need to keep the plugin
+       record in tact so that the failure information is kept
+       around. */
+    if (p->state < KMM_PLUGIN_STATE_NONE) {
+        release_plugin = FALSE;
+    }
+
     LeaveCriticalSection(&cs_kmm);
 
     if(!np) {
@@ -489,8 +558,9 @@ void kmm_exit_plugin(kmm_plugin_i * p) {
         kmq_post_message(KMSG_KMM, KMSG_KMM_I_REG, KMM_REG_EXIT_MODULE, (void *) p->module);
     }
 
-    /* release the hold obtained in kmm_init_plugin() */
-    kmm_release_plugin(kmm_handle_from_plugin(p));
+    /* release the hold obtained in kmmint_init_plugin() */
+    if (release_plugin)
+        kmm_release_plugin(kmm_handle_from_plugin(p));
 }
 
 /*! \internal
@@ -499,7 +569,7 @@ void kmm_exit_plugin(kmm_plugin_i * p) {
   \a m is not in the linked list yet.
 
   \note Should only be called from the context of the registrar thread. */
-void kmm_init_module(kmm_module_i * m) {
+void kmmint_init_module(kmm_module_i * m) {
     HMODULE hm;
     init_module_t p_init_module;
     kmm_plugin_i * pi;
@@ -538,7 +608,7 @@ void kmm_init_module(kmm_module_i * m) {
     /* If the module is not in the pre-init state, we can't
        initialize it. */
     if(m->state != KMM_MODULE_STATE_PREINIT) {
-        _report_mr1(KHERR_WARNING, MSG_IM_NOT_PREINIT, _int32(m->state));
+        _report_mr1(KHERR_INFO, MSG_IM_NOT_PREINIT, _int32(m->state));
         goto _exit;
     }
 
@@ -549,48 +619,56 @@ void kmm_init_module(kmm_module_i * m) {
         goto _exit;
     }
 
-    if(KHM_SUCCEEDED(khc_read_int32(csp_mod, L"Flags", &i))) {
-        if(i & KMM_MODULE_FLAG_DISABLED) {
-            _report_mr0(KHERR_ERROR, MSG_IM_DISABLED);
+    if(KHM_SUCCEEDED(khc_read_int32(csp_mod, L"Flags", &i)) &&
+       (i & KMM_MODULE_FLAG_DISABLED)) {
 
-            m->state = KMM_MODULE_STATE_FAIL_DISABLED;
-            goto _exit;
-        }
+        _report_mr0(KHERR_INFO, MSG_IM_DISABLED);
+
+        m->state = KMM_MODULE_STATE_FAIL_DISABLED;
+        goto _exit;
     }
 
     if(KHM_SUCCEEDED(khc_read_int32(csp_mod, L"FailureCount", &i))) {
         khm_int64 tm;
         khm_int64 ct;
+        FILETIME fct;
+        khm_int32 last_reason = 0;
 
         /* reset the failure count if the failure count reset time
            period has elapsed */
         tm = 0;
         khc_read_int64(csp_mod, L"FailureTime", &tm);
-        GetSystemTimeAsFileTime((LPFILETIME) &ct);
-        ct -= tm;
+        GetSystemTimeAsFileTime(&fct);
+
+        ct = (FtToInt(&fct) - tm) / 10000000i64;
 
         if(tm > 0 && 
-           FtIntervalToSeconds((LPFILETIME) &ct) > fail_reset_time) {
-
+           ct > fail_reset_time) {
             i = 0;
             khc_write_int32(csp_mod, L"FailureCount", 0);
             khc_write_int64(csp_mod, L"FailureTime", 0);
-
         }
 
-        if(i > max_fail_count) {
+        khc_read_int32(csp_mod, L"FailureReason", &last_reason);
+
+        /* did we exceed the max failure count?  However, we ignore
+           the max failure count if the reason why it didn't load the
+           last time was because the module wasn't found. */
+        if(i > max_fail_count && 
+           last_reason != KMM_MODULE_STATE_FAIL_NOT_FOUND) {
             /* failed too many times */
-            _report_mr0(KHERR_ERROR, MSG_IM_MAX_FAIL);
+            _report_mr0(KHERR_INFO, MSG_IM_MAX_FAIL);
 
             m->state = KMM_MODULE_STATE_FAIL_MAX_FAILURE;
             goto _exit;
         }
     }
 
-    if(khc_read_string(csp_mod, L"ImagePath", NULL, &sz) == KHM_ERROR_TOO_LONG) {
+    if(khc_read_string(csp_mod, L"ImagePath", NULL, &sz) == 
+       KHM_ERROR_TOO_LONG) {
         if(m->path)
-            free(m->path);
-        m->path = malloc(sz);
+            PFREE(m->path);
+        m->path = PMALLOC(sz);
         khc_read_string(csp_mod, L"ImagePath", m->path, &sz);
     } else {
         _report_mr0(KHERR_ERROR, MSG_IM_NOT_REGISTERED);
@@ -599,11 +677,23 @@ void kmm_init_module(kmm_module_i * m) {
         goto _exit;
     }
 
-    if (khc_read_string(csp_mod, L"Vendor", NULL, &sz) == KHM_ERROR_TOO_LONG) {
-        if (m->vendor)
-            free(m->vendor);
-        m->vendor = malloc(sz);
-        khc_read_string(csp_mod, L"Vendor", m->vendor, &sz);
+    rv = kmmint_read_module_info(m);
+
+    if (KHM_FAILED(rv)) {
+        if (rv == KHM_ERROR_INCOMPATIBLE) {
+            _report_mr0(KHERR_ERROR, MSG_IM_INCOMPATIBLE);
+
+            m->state = KMM_MODULE_STATE_FAIL_INCOMPAT;
+        } else if (rv == KHM_ERROR_NOT_FOUND) {
+            _report_mr1(KHERR_ERROR, MSG_IM_NOT_FOUND, _dupstr(m->path));
+
+            m->state = KMM_MODULE_STATE_FAIL_NOT_FOUND;
+        } else {
+            _report_mr0(KHERR_ERROR, MSG_IM_INVALID_MODULE);
+
+            m->state = KMM_MODULE_STATE_FAIL_INV_MODULE;
+        }
+        goto _exit;
     }
 
     /* check again */
@@ -613,11 +703,13 @@ void kmm_init_module(kmm_module_i * m) {
         goto _exit;
     }
 
+    /* from this point on, we must record any failure codes */
+    record_failure = TRUE;
+
     hm = LoadLibrary(m->path);
     if(!hm) {
         m->h_module = NULL;
         m->state = KMM_MODULE_STATE_FAIL_NOT_FOUND;
-        record_failure = TRUE;
 
         _report_mr1(KHERR_ERROR, MSG_IM_NOT_FOUND, _dupstr(m->path));
 
@@ -626,14 +718,17 @@ void kmm_init_module(kmm_module_i * m) {
 
     /* from this point on, we need to discard the module through
        exit_module */
+    ResetEvent(evt_exit);
+
+    kmm_active_modules++;
+
     release_module = FALSE;
     exit_module = TRUE;
-    record_failure = TRUE;
 
     m->flags |= KMM_MODULE_FLAG_LOADED;
     m->h_module = hm;
 
-    /*TODO: check signatures */
+    /* TODO: check signatures */
 
     p_init_module = (init_module_t) GetProcAddress(hm, EXP_INIT_MODULE);
 
@@ -646,7 +741,6 @@ void kmm_init_module(kmm_module_i * m) {
 
     m->state = KMM_MODULE_STATE_INIT;
 
-
     /* call init_module() */
     rv = (*p_init_module)(kmm_handle_from_module(m));
 
@@ -673,7 +767,7 @@ void kmm_init_module(kmm_module_i * m) {
         LPOP(&(m->plugins), &pi);
         if(pi) {
             pi->flags &= ~KMM_PLUGIN_FLAG_IN_MODLIST;
-            kmm_init_plugin(pi);
+            kmmint_init_plugin(pi);
 
             /* release the hold obtained in kmm_provide_plugin() */
             kmm_release_plugin(kmm_handle_from_plugin(pi));
@@ -681,9 +775,14 @@ void kmm_init_module(kmm_module_i * m) {
     } while(pi);
 
     if(!m->plugin_count) {
+        /* We don't want to report this case.  This usually means that
+           the plugins that were provided by the module were
+           disabled. */
+#ifdef REPORT_EMPTY_MODULES
         _report_mr0(KHERR_ERROR, MSG_IM_NO_PLUGINS);
 
         m->state = KMM_MODULE_STATE_FAIL_NO_PLUGINS;
+#endif
         record_failure = FALSE;
         goto _exit;
     }
@@ -693,12 +792,10 @@ void kmm_init_module(kmm_module_i * m) {
     exit_module = FALSE;
     record_failure = FALSE;
 
-    ResetEvent(evt_exit);
-
-_exit:
+ _exit:
     if(csp_mod) {
         if(record_failure) {
-            khm_int64 ct;
+            FILETIME fct;
 
             i = 0;
             khc_read_int32(csp_mod, L"FailureCount", &i);
@@ -706,27 +803,74 @@ _exit:
             khc_write_int32(csp_mod, L"FailureCount", i);
 
             if(i==1) { /* first fault */
-                GetSystemTimeAsFileTime((LPFILETIME) &ct);
-                khc_write_int64(csp_mod, L"FailureTime", ct);
+                GetSystemTimeAsFileTime(&fct);
+                khc_write_int64(csp_mod, L"FailureTime", FtToInt(&fct));
             }
+
+            khc_write_int32(csp_mod, L"FailureReason", m->state);
         }
         khc_close_space(csp_mod);
     }
+
     if(csp_mods)
         khc_close_space(csp_mods);
 
     _report_mr2(KHERR_INFO, MSG_IM_MOD_STATE, 
                 _dupstr(m->name), _int32(m->state));
 
-    if(release_module)
-        kmm_release_module(kmm_handle_from_module(m));
-
     kmmint_remove_from_module_queue();
 
     /* if something went wrong after init_module was called on the
        module code, we need to call exit_module */
     if(exit_module)
-        kmm_exit_module(m);
+        kmmint_exit_module(m);
+
+    if(release_module)
+        kmm_release_module(kmm_handle_from_module(m));
+
+    if (kherr_is_error()) {
+        kherr_context * c;
+        kherr_event * err_e = NULL;
+        kherr_event * warn_e = NULL;
+        kherr_event * e;
+
+        c = kherr_peek_context();
+        err_e = kherr_get_err_event(c);
+        for(e = kherr_get_first_event(c);
+            e;
+            e = kherr_get_next_event(e)) {
+            if (e != err_e &&
+                e->severity == KHERR_WARNING) {
+                warn_e = e;
+                break;
+            }
+        }
+
+        kherr_evaluate_event(err_e);
+        if (warn_e)
+            kherr_evaluate_event(warn_e);
+
+        kherr_clear_error();
+
+        e = kherr_report(KHERR_ERROR,
+                         (wchar_t *) MSG_IMERR_TITLE,
+                         KHERR_FACILITY,
+                         NULL,
+                         err_e->long_desc,
+                         ((warn_e)? (wchar_t *)MSG_IMERR_SUGGEST: NULL),
+                         KHERR_FACILITY_ID,
+                         KHERR_SUGGEST_NONE,
+                         _cstr(m->name),
+                         ((warn_e)? _cstr(warn_e->long_desc):_vnull()),
+                         _vnull(),_vnull(),
+                         KHERR_RF_MSG_SHORT_DESC |
+                         ((warn_e)? KHERR_RF_MSG_SUGGEST: 0),
+                         KHERR_HMODULE);
+
+        kherr_evaluate_event(e);
+
+        kherr_release_context(c);
+    }
 
     _end_task();
 }
@@ -737,21 +881,21 @@ _exit:
 
   \note Should only be called from the context of the registrar
   thread */
-void kmm_exit_module(kmm_module_i * m) {
+void kmmint_exit_module(kmm_module_i * m) {
     kmm_plugin_i * p;
 
-    /*  exiting a module happens in two stages.  
-    
+    /*  Exiting a module happens in two stages.  
+
         If the module state is running (there are active plugins) then
         those plugins must be exited.  This has to be done from the
         plugin threads.  The signal for the plugins to exit must be
         issued from the registrar.  Therefore, we post messages to the
         registrar for each plugin we want to remove and exit
-        kmm_exit_module().
+        kmmint_exit_module().
 
         When the last plugin is exited, the plugin management code
         automatically signalls the registrar to remove the module.
-        kmm_exit_module() gets called again.  This is the second
+        kmmint_exit_module() gets called again.  This is the second
         stage, where we call exit_module() for the module and start
         unloading everything.
     */
@@ -762,7 +906,7 @@ void kmm_exit_module(kmm_module_i * m) {
     LPOP(&(m->plugins), &p);
     while(p) {
         p->flags &= ~KMM_PLUGIN_FLAG_IN_MODLIST;
-        kmm_exit_plugin(p);
+        kmmint_exit_plugin(p);
 
         /* release hold from kmm_provide_plugin() */
         kmm_release_plugin(kmm_handle_from_plugin(p));
@@ -780,7 +924,8 @@ void kmm_exit_module(kmm_module_i * m) {
         while(p) {
             if(p->module == m) {
                 kmm_hold_plugin(kmm_handle_from_plugin(p));
-                kmq_post_message(KMSG_KMM, KMSG_KMM_I_REG, KMM_REG_EXIT_PLUGIN, (void *) p);
+                kmq_post_message(KMSG_KMM, KMSG_KMM_I_REG, 
+                                 KMM_REG_EXIT_PLUGIN, (void *) p);
                 np++;
             }
 
@@ -797,8 +942,7 @@ void kmm_exit_module(kmm_module_i * m) {
         }
     }
 
-    if(m->flags & KMM_MODULE_FLAG_INITP)
-    {
+    if(m->flags & KMM_MODULE_FLAG_INITP) {
         exit_module_t p_exit_module;
 
         if(m->state > 0)
@@ -809,16 +953,16 @@ void kmm_exit_module(kmm_module_i * m) {
                                            EXP_EXIT_MODULE);
         if(p_exit_module) {
             LeaveCriticalSection(&cs_kmm);
-            p_exit_module(kmm_handle_from_module(m));
+            (*p_exit_module)(kmm_handle_from_module(m));
             EnterCriticalSection(&cs_kmm);
         }
     }
 
-    LeaveCriticalSection(&cs_kmm);
-
     if(m->state > 0)
         m->state = KMM_MODULE_STATE_EXITED;
 
+    LeaveCriticalSection(&cs_kmm);
+
     if(m->h_module) {
         FreeLibrary(m->h_module);
     }
@@ -831,6 +975,19 @@ void kmm_exit_module(kmm_module_i * m) {
     m->h_resource = NULL;
     m->flags = 0;
 
-    /* release the hold obtained in kmm_init_module() */
+    /* release the hold obtained in kmmint_init_module() */
     kmm_release_module(kmm_handle_from_module(m));
+
+    /* Last but not least, now see if there are any modules left that
+       are running. If not, we can safely signal an exit. */
+
+#ifdef DEBUG
+    assert(kmm_active_modules > 0);
+#endif
+
+    kmm_active_modules--;
+
+    if (kmm_active_modules == 0) {
+        SetEvent(evt_exit);
+    }
 }
