@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2004 Massachusetts Institute of Technology
+ * Copyright (c) 2005 Massachusetts Institute of Technology
  *
  * Permission is hereby granted, free of charge, to any person
  * obtaining a copy of this software and associated documentation
@@ -43,12 +43,20 @@ nc_common_dlg_proc(HWND hwnd,
 {
     switch(uMsg) {
     case WM_INITDIALOG:
+        {
+            khui_nc_wnd_data * d;
+
+            d = (khui_nc_wnd_data *) lParam;
 
 #pragma warning(push)
 #pragma warning(disable: 4244)
-        SetWindowLongPtr(hwnd, DWLP_USER, lParam);
+            SetWindowLongPtr(hwnd, DWLP_USER, lParam);
 #pragma warning(pop)
-
+            if (d->nc->subtype == KMSG_CRED_PASSWORD) {
+                ShowWindow(GetDlgItem(hwnd, IDC_NC_OPTIONS),
+                           SW_HIDE);
+            }
+        }
         return TRUE;
 
     case WM_COMMAND:
@@ -244,7 +252,7 @@ nc_update_credtext(khui_nc_wnd_data * d)
     HWND hw = NULL;
     size_t cch = 0;
 
-    ctbuf = malloc(NC_MAXCB_CREDTEXT);
+    ctbuf = PMALLOC(NC_MAXCB_CREDTEXT);
 
     assert(ctbuf != NULL);
 
@@ -271,6 +279,7 @@ nc_update_credtext(khui_nc_wnd_data * d)
         kcdb_identity_get_name(d->nc->identities[0], id_name, &cbbuf);
 
         kcdb_identity_get_flags(d->nc->identities[0], &flags);
+
         if (flags & KCDB_IDENT_FLAG_INVALID) {
             LoadString(khm_hInstance, IDS_NC_CREDTEXT_ID_INVALID, 
                        id_fmt, (int) ARRAYLENGTH(id_fmt));
@@ -290,6 +299,21 @@ nc_update_credtext(khui_nc_wnd_data * d)
         StringCbPrintf(buf, NC_MAXCB_CREDTEXT - cch*sizeof(wchar_t), 
                        main_fmt, id_string);
 
+        if (flags & KCDB_IDENT_FLAG_VALID) {
+            if (flags & KCDB_IDENT_FLAG_DEFAULT)
+                LoadString(khm_hInstance, IDS_NC_ID_DEF,
+                           id_string, ARRAYLENGTH(id_string));
+            else if (d->nc->set_default)
+                LoadString(khm_hInstance, IDS_NC_ID_WDEF,
+                           id_string, ARRAYLENGTH(id_string));
+            else
+                LoadString(khm_hInstance, IDS_NC_ID_NDEF,
+                           id_string, ARRAYLENGTH(id_string));
+
+            StringCbCat(buf, NC_MAXCB_CREDTEXT - cch * sizeof(wchar_t),
+                        id_string);
+        }
+
     } else if(d->nc->n_identities > 1) {
         wchar_t *ids_string;
         khm_size cb_ids_string;
@@ -308,7 +332,7 @@ nc_update_credtext(khui_nc_wnd_data * d)
            a comma separated string */
 
         /* d->nc->n_identities is at least 2 */
-        ids_string = malloc((KCDB_IDENT_MAXCB_NAME + sizeof(id_fmt)) * 
+        ids_string = PMALLOC((KCDB_IDENT_MAXCB_NAME + sizeof(id_fmt)) * 
                             (d->nc->n_identities - 1));
         cb_ids_string = 
             (KCDB_IDENT_MAXCB_NAME + sizeof(id_fmt)) * 
@@ -365,7 +389,7 @@ nc_update_credtext(khui_nc_wnd_data * d)
             StringCbPrintf(buf, NC_MAXCB_CREDTEXT - cch*sizeof(wchar_t), 
                            main_fmt, id_string, ids_string);
 
-            free(ids_string);
+            PFREE(ids_string);
         }
     } else {
         LoadString(khm_hInstance, IDS_NC_CREDTEXT_ID_NONE, 
@@ -393,7 +417,7 @@ nc_update_credtext(khui_nc_wnd_data * d)
 
     SetDlgItemText(d->dlg_main, IDC_NC_CREDTEXT, ctbuf);
 
-    free(ctbuf);
+    PFREE(ctbuf);
 
     /* so depending on whether the primary identity was found to be
        invalid, we need to disable the Ok button and set the title to
@@ -456,6 +480,12 @@ nc_update_credtext(khui_nc_wnd_data * d)
 
 #define CW_PARAM DWLP_USER
 
+static void
+nc_add_control_row(khui_nc_wnd_data * d, 
+                   HWND label,
+                   HWND input,
+                   khui_control_size size);
+
 static LRESULT 
 nc_handle_wm_create(HWND hwnd,
                     UINT uMsg,
@@ -471,7 +501,7 @@ nc_handle_wm_create(HWND hwnd,
 
     lpc = (LPCREATESTRUCT) lParam;
 
-    ncd = malloc(sizeof(*ncd));
+    ncd = PMALLOC(sizeof(*ncd));
     ZeroMemory(ncd, sizeof(*ncd));
 
     c = (khui_new_creds *) lpc->lpCreateParams;
@@ -691,6 +721,7 @@ nc_handle_wm_create(HWND hwnd,
         ShowWindow(ncd->dlg_ts, SW_HIDE);
 
         nc_position_credtext(ncd);
+
     } else {
         /* hide and show stuff */
         ShowWindow(ncd->dlg_main, SW_SHOW);
@@ -705,6 +736,28 @@ nc_handle_wm_create(HWND hwnd,
        selector controls */
     c->ident_cb(c, WMNC_IDENT_INIT, NULL, 0, 0, (LPARAM) ncd->dlg_main);
 
+#if 0
+    {
+        HWND hw;
+        wchar_t wcaption[64];
+
+        LoadString(khm_hInstance, IDS_NC_SETDEF, wcaption,
+                   ARRAYLENGTH(wcaption));
+
+        /* Now create the set as default button */
+        hw = CreateWindow
+            (L"BUTTON",
+             wcaption,
+             WS_CHILD | WS_VISIBLE | WS_TABSTOP | BS_AUTOCHECKBOX,
+             0, 0, 100, 100,
+             ncd->dlg_main,
+             (HMENU) NC_BN_SET_DEF_ID,
+             khm_hInstance,
+             NULL);
+
+        nc_add_control_row(ncd, NULL, hw, KHUI_CTRLSIZE_HALF);
+    }
+#endif
     /* we defer the creation of the tab buttons for later */
 
     /* add this to the dialog chain */
@@ -750,6 +803,8 @@ nc_add_control_row(khui_nc_wnd_data * d,
         r_row.bottom += r_input.bottom;
         OffsetRect(&r_label, r_row.left, r_row.top);
     } else {
+        SetRectEmpty(&r_label);
+        SetRectEmpty(&r_input);
 #ifdef DEBUG
         assert(FALSE);
 #else
@@ -757,23 +812,27 @@ nc_add_control_row(khui_nc_wnd_data * d,
 #endif
     }
 
-    SetWindowPos(label,
-                 ((d->hwnd_last_idspec != NULL)?
-                  d->hwnd_last_idspec:
-                  HWND_TOP),
-                 r_label.left, r_label.top,
-                 r_label.right - r_label.left,
-                 r_label.bottom - r_label.top,
-                 SWP_DEFERERASE | SWP_NOACTIVATE |
-                 SWP_NOOWNERZORDER);
+    if (label)
+        SetWindowPos(label,
+                     ((d->hwnd_last_idspec != NULL)?
+                      d->hwnd_last_idspec:
+                      HWND_TOP),
+                     r_label.left, r_label.top,
+                     r_label.right - r_label.left,
+                     r_label.bottom - r_label.top,
+                     SWP_DEFERERASE | SWP_NOACTIVATE |
+                     SWP_NOOWNERZORDER);
 
-    SetWindowPos(input,
-                 label,
-                 r_input.left, r_input.top,
-                 r_input.right - r_input.left,
-                 r_input.bottom - r_input.top,
-                 SWP_DEFERERASE | SWP_NOACTIVATE |
-                 SWP_NOOWNERZORDER);
+    if (input)
+        SetWindowPos(input,
+                     (label ? label : ((d->hwnd_last_idspec != NULL)?
+                                       d->hwnd_last_idspec:
+                                       HWND_TOP)),
+                     r_input.left, r_input.top,
+                     r_input.right - r_input.left,
+                     r_input.bottom - r_input.top,
+                     SWP_DEFERERASE | SWP_NOACTIVATE |
+                     SWP_NOOWNERZORDER);
 
     d->hwnd_last_idspec = input;
 
@@ -821,7 +880,7 @@ nc_handle_wm_destroy(HWND hwnd,
     d->dlg_main = NULL;
     d->dlg_ts = NULL;
 
-    free(d);
+    PFREE(d);
 
     return TRUE;
 }
@@ -842,7 +901,7 @@ nc_handle_wm_command(HWND hwnd,
         switch(LOWORD(wParam)) {
 
         case IDOK:
-            d->nc->result = KHUI_NC_RESULT_GET_CREDS;
+            d->nc->result = KHUI_NC_RESULT_PROCESS;
 
             /* fallthrough */
 
@@ -877,6 +936,10 @@ nc_handle_wm_command(HWND hwnd,
             }
             return FALSE;
 
+        case IDC_NC_HELP:
+            khm_html_help(hwnd, NULL, HH_HELP_CONTEXT, IDH_ACTION_NEW_ID);
+            return FALSE;
+
         case IDC_NC_OPTIONS: 
             /* the Options button in the main window was clicked.  we
                respond by expanding the dialog. */
@@ -958,10 +1021,26 @@ nc_handle_wm_command(HWND hwnd,
                                                    WMNC_DIALOG_SWITCH_PANEL),
                                         0);
                     }
+                } else if (!wcsicmp(sid, L"NotDef")) {
+                    d->nc->set_default = FALSE;
+                    nc_update_credtext(d);
+                } else if (!wcsicmp(sid, L"MakeDef")) {
+                    d->nc->set_default = TRUE;
+                    nc_update_credtext(d);
                 }
             }
             return FALSE;
 
+#if 0
+        case NC_BN_SET_DEF_ID:
+            {
+                d->nc->set_default =
+                    (IsDlgButtonChecked(d->dlg_main, NC_BN_SET_DEF_ID)
+                     == BST_CHECKED);
+            }
+            return FALSE;
+#endif
+
         default:
             /* if one of the tab strip buttons were pressed, then
                we should switch to that panel */
@@ -1101,7 +1180,7 @@ static LRESULT nc_handle_wm_nc_notify(HWND hwnd,
                 if (d->nc->types[i]->dlg_proc == NULL) {
                     d->nc->types[i]->hwnd_panel = NULL;
                 } else {
-                    /* Create the dialog panel */ 
+                    /* Create the dialog panel */
                     d->nc->types[i]->hwnd_panel = 
                         CreateDialogParam(d->nc->types[i]->h_module,
                                           d->nc->types[i]->dlg_template,
@@ -1173,7 +1252,7 @@ static LRESULT nc_handle_wm_nc_notify(HWND hwnd,
                       nc_tab_sort_func);
 
                 for(i=0; i < d->nc->n_types;i++) {
-                    wchar_t * name;
+                    wchar_t * name = NULL;
 
                     d->nc->types[i]->ordinal = i + 1;
 
@@ -1188,7 +1267,7 @@ static LRESULT nc_handle_wm_nc_notify(HWND hwnd,
                             &cbsize, 
                             KCDB_TS_SHORT) == KHM_ERROR_TOO_LONG) {
 
-                            name = malloc(cbsize);
+                            name = PMALLOC(cbsize);
                             kcdb_credtype_describe(d->nc->types[i]->type, 
                                                    name, 
                                                    &cbsize, 
@@ -1235,7 +1314,7 @@ static LRESULT nc_handle_wm_nc_notify(HWND hwnd,
                     x += width;
 
                     if(!(d->nc->types[i]->name))
-                        free(name);
+                        PFREE(name);
 
                     /* Now set the position of the type panel */
                     ShowWindow(d->nc->types[i]->hwnd_panel, SW_HIDE);
@@ -1271,7 +1350,21 @@ static LRESULT nc_handle_wm_nc_notify(HWND hwnd,
 
             nc_notify_types(d->nc, KHUI_WM_NC_NOTIFY,
                             MAKEWPARAM(0, WMNC_IDENTITY_CHANGE), 0);
+
+            if (d->nc->subtype == KMSG_CRED_NEW_CREDS &&
+                d->nc->n_identities > 0 &&
+                d->nc->identities[0]) {
+                khm_int32 f = 0;
+
+                kcdb_identity_get_flags(d->nc->identities[0], &f);
+
+                if (!(f & KCDB_IDENT_FLAG_DEFAULT)) {
+                    d->nc->set_default = FALSE;
+                }
+            }
+
             nc_update_credtext(d);
+
         }
         break;
 
@@ -1517,7 +1610,11 @@ static LRESULT nc_handle_wm_nc_notify(HWND hwnd,
                              SWP_NOACTIVATE | SWP_NOMOVE | 
                              SWP_NOOWNERZORDER | SWP_NOSIZE | 
                              SWP_SHOWWINDOW);
-                                
+
+                SendMessage(hw, EM_SETLIMITTEXT,
+                            KHUI_MAXCCH_PROMPT_VALUE -1,
+                            0);
+
                 d->nc->prompts[i]->hwnd_edit = hw;
 
                 hw_prev = hw;
@@ -1542,12 +1639,12 @@ static LRESULT nc_handle_wm_nc_notify(HWND hwnd,
 
             nc = d->nc;
 
-            /* reset state */
-            nc->result = KHUI_NC_RESULT_CANCEL;
-
             if(nc->response & KHUI_NC_RESPONSE_NOEXIT) {
                 HWND hw;
 
+                /* reset state */
+                nc->result = KHUI_NC_RESULT_CANCEL;
+
                 hw = GetDlgItem(d->dlg_main, IDOK);
                 EnableWindow(hw, TRUE);
                 hw = GetDlgItem(d->dlg_main, IDCANCEL);
@@ -1586,6 +1683,72 @@ static LRESULT nc_handle_wm_nc_notify(HWND hwnd,
     return TRUE;
 }
 
+static LRESULT nc_handle_wm_help(HWND hwnd,
+                                 UINT uMsg,
+                                 WPARAM wParam,
+                                 LPARAM lParam) {
+    static DWORD ctxids[] = {
+        NC_TS_CTRL_ID_MIN, IDH_NC_TABMAIN,
+        NC_TS_CTRL_ID_MIN + 1, IDH_NC_TABBUTTON,
+        NC_TS_CTRL_ID_MIN + 2, IDH_NC_TABBUTTON,
+        NC_TS_CTRL_ID_MIN + 3, IDH_NC_TABBUTTON,
+        NC_TS_CTRL_ID_MIN + 4, IDH_NC_TABBUTTON,
+        NC_TS_CTRL_ID_MIN + 5, IDH_NC_TABBUTTON,
+        NC_TS_CTRL_ID_MIN + 6, IDH_NC_TABBUTTON,
+        NC_TS_CTRL_ID_MIN + 7, IDH_NC_TABBUTTON,
+        IDOK, IDH_NC_OK,
+        IDCANCEL, IDH_NC_CANCEL,
+        IDC_NC_HELP, IDH_NC_HELP,
+        IDC_NC_OPTIONS, IDH_NC_OPTIONS,
+        IDC_NC_CREDTEXT, IDH_NC_CREDWND,
+        0
+    };
+
+    HELPINFO * hlp;
+    HWND hw = NULL;
+    HWND hw_ctrl;
+    khui_nc_wnd_data * d;
+
+    d = (khui_nc_wnd_data *)(LONG_PTR) GetWindowLongPtr(hwnd, CW_PARAM);
+
+    hlp = (HELPINFO *) lParam;
+
+    if (d->nc->subtype != KMSG_CRED_NEW_CREDS &&
+        d->nc->subtype != KMSG_CRED_PASSWORD)
+        return TRUE;
+
+    if (hlp->iContextType != HELPINFO_WINDOW)
+        return TRUE;
+
+    if (hlp->hItemHandle != NULL &&
+        hlp->hItemHandle != hwnd) {
+        DWORD id;
+        int i;
+
+        hw_ctrl =hlp->hItemHandle;
+
+        id = GetWindowLong(hw_ctrl, GWL_ID);
+        for (i=0; ctxids[i] != 0; i += 2)
+            if (ctxids[i] == id)
+                break;
+
+        if (ctxids[i] != 0)
+            hw = khm_html_help(hw_ctrl,
+                               ((d->nc->subtype == KMSG_CRED_NEW_CREDS)?
+                                L"::popups_newcreds.txt":
+                                L"::popups_password.txt"),
+                               HH_TP_HELP_WM_HELP,
+                               (DWORD_PTR) ctxids);
+    }
+
+    if (hw == NULL) {
+        khm_html_help(hwnd, NULL, HH_HELP_CONTEXT,
+                      ((d->nc->subtype == KMSG_CRED_NEW_CREDS)?
+                       IDH_ACTION_NEW_ID: IDH_ACTION_PASSWD_ID));
+    }
+
+    return TRUE;
+}
 
 static LRESULT CALLBACK nc_window_proc(HWND hwnd,
                                        UINT uMsg,
@@ -1606,6 +1769,9 @@ static LRESULT CALLBACK nc_window_proc(HWND hwnd,
     case WM_MOVING:
         return nc_handle_wm_moving(hwnd, uMsg, wParam, lParam);
 
+    case WM_HELP:
+        return nc_handle_wm_help(hwnd, uMsg, wParam, lParam);
+
     case KHUI_WM_NC_NOTIFY:
         return nc_handle_wm_nc_notify(hwnd, uMsg, wParam, lParam);
     }
