@@ -1,5 +1,5 @@
 /*
-Copyright 2005,2006 by the Massachusetts Institute of Technology
+Copyright 2005 by the Massachusetts Institute of Technology
 
 All rights reserved.
 
@@ -24,9 +24,6 @@ SOFTWARE.
 
 #include "kfwlogon.h"
 #include <winbase.h>
-#include <Aclapi.h>
-#include <userenv.h>
-#include <Sddl.h>
 
 #include <io.h>
 #include <sys/stat.h>
@@ -761,152 +758,19 @@ KFW_get_cred( char * username,
     return(code);
 }
 
-int KFW_set_ccache_dacl(char *filename, HANDLE hUserToken)
-{
-    // SID_IDENTIFIER_AUTHORITY authority = SECURITY_NT_SID_AUTHORITY;
-    PSID pSystemSID = NULL;
-    DWORD SystemSIDlength = 0, UserSIDlength = 0;
-    PACL ccacheACL = NULL;
-    DWORD ccacheACLlength = 0;
-    PTOKEN_USER pTokenUser = NULL;
-    DWORD retLen;
-    DWORD gle;
-    int ret = 0;  
-
-    if (!filename) {
-	DebugEvent0("KFW_set_ccache_dacl - invalid parms");
-	return 1;
-    }
-
-    /* Get System SID */
-    if (!ConvertStringSidToSid("S-1-5-18", &pSystemSID)) {
-	DebugEvent("KFW_set_ccache_dacl - ConvertStringSidToSid GLE = 0x%x", GetLastError());
-	ret = 1;
-	goto cleanup;
-    }
-
-    /* Create ACL */
-    SystemSIDlength = GetLengthSid(pSystemSID);
-    ccacheACLlength = sizeof(ACL) + sizeof(ACCESS_ALLOWED_ACE)
-        + SystemSIDlength - sizeof(DWORD);
-
-    if (hUserToken) {
-	if (!GetTokenInformation(hUserToken, TokenUser, NULL, 0, &retLen))
-	{
-	    if ( GetLastError() == ERROR_INSUFFICIENT_BUFFER ) {
-		pTokenUser = (PTOKEN_USER) LocalAlloc(LPTR, retLen);
-
-		if (!GetTokenInformation(hUserToken, TokenUser, pTokenUser, retLen, &retLen))
-		{
-		    DebugEvent("GetTokenInformation failed: GLE = %lX", GetLastError());
-		}
-	    }		 
-	}
-
-	if (pTokenUser) {
-	    UserSIDlength = GetLengthSid(pTokenUser->User.Sid);
-
-	    ccacheACLlength += sizeof(ACCESS_ALLOWED_ACE) + UserSIDlength 
-		- sizeof(DWORD);
-	}
-    }
-
-    ccacheACL = (PACL) LocalAlloc(LPTR, ccacheACLlength);
-    if (!ccacheACL) {
-	DebugEvent("KFW_set_ccache_dacl - LocalAlloc GLE = 0x%x", GetLastError());
-	ret = 1;
-	goto cleanup;
-    }
-
-    InitializeAcl(ccacheACL, ccacheACLlength, ACL_REVISION);
-    AddAccessAllowedAceEx(ccacheACL, ACL_REVISION, 0,
-                         STANDARD_RIGHTS_ALL | SPECIFIC_RIGHTS_ALL,
-                         pSystemSID);
-    if (pTokenUser) {
-	AddAccessAllowedAceEx(ccacheACL, ACL_REVISION, 0,
-			     STANDARD_RIGHTS_ALL | SPECIFIC_RIGHTS_ALL,
-			     pTokenUser->User.Sid);
-	if (!SetNamedSecurityInfo( filename, SE_FILE_OBJECT,
-				   DACL_SECURITY_INFORMATION | PROTECTED_DACL_SECURITY_INFORMATION,
-				   NULL,
-				   NULL, 
-				   ccacheACL,
-				   NULL)) {
-	    gle = GetLastError();
-	    DebugEvent("SetNamedSecurityInfo DACL failed: GLE = 0x%lX", gle);
-	    if (gle != ERROR_NO_TOKEN)
-		ret = 1;
-	}
-	if (!SetNamedSecurityInfo( filename, SE_FILE_OBJECT,
-				   OWNER_SECURITY_INFORMATION,
-				   pTokenUser->User.Sid,
-				   NULL, 
-				   NULL,
-				   NULL)) {
-	    gle = GetLastError();
-	    DebugEvent("SetNamedSecurityInfo DACL failed: GLE = 0x%lX", gle);
-	    if (gle != ERROR_NO_TOKEN)
-		ret = 1;
-	}
-    } else {
-	if (!SetNamedSecurityInfo( filename, SE_FILE_OBJECT,
-				   DACL_SECURITY_INFORMATION | PROTECTED_DACL_SECURITY_INFORMATION,
-				   NULL,
-				   NULL, 
-				   ccacheACL,
-				   NULL)) {
-	    gle = GetLastError();
-	    DebugEvent("SetNamedSecurityInfo DACL failed: GLE = 0x%lX", gle);
-	    if (gle != ERROR_NO_TOKEN)
-		ret = 1;
-	}
-    }
-
-  cleanup:
-    if (pSystemSID)
-	LocalFree(pSystemSID);
-    if (pTokenUser)
-	LocalFree(pTokenUser);
-    if (ccacheACL)
-	LocalFree(ccacheACL);
-    return ret;
-}
-
-int KFW_obtain_user_temp_directory(HANDLE hUserToken, char *newfilename, int size)
-{
-    int  retval = 0;
-    DWORD dwSize = size-1;	/* leave room for nul */
-    DWORD dwLen  = 0;
-
-    if (!hUserToken || !newfilename || size <= 0)
-	return;
-
-    *newfilename = '\0';
-
-    dwLen = ExpandEnvironmentStringsForUser(hUserToken, "%TEMP%", newfilename, dwSize);
-    if ( !dwLen || dwLen > dwSize )
-	dwLen = ExpandEnvironmentStringsForUser(hUserToken, "%TMP%", newfilename, dwSize);
-    if ( !dwLen || dwLen > dwSize )
-	return 1;
-
-    newfilename[dwSize] = '\0';
-    return 0;
-}
-
 void
 KFW_copy_cache_to_system_file(char * user, char * szLogonId)
 {
-    char filename[MAX_PATH] = "";
+    char filename[256];
     DWORD count;
-    char cachename[MAX_PATH + 8] = "FILE:";
+    char cachename[264] = "FILE:";
     krb5_context		ctx = 0;
     krb5_error_code		code;
     krb5_principal              princ = 0;
     krb5_ccache			cc  = 0;
     krb5_ccache                 ncc = 0;
-    PSECURITY_ATTRIBUTES        pSA = NULL;
-    
-    if (!pkrb5_init_context || !user || !szLogonId)
+
+    if (!pkrb5_init_context)
         return;
 
     count = GetEnvironmentVariable("TEMP", filename, sizeof(filename));
@@ -925,8 +789,6 @@ KFW_copy_cache_to_system_file(char * user, char * szLogonId)
 
     strcat(cachename, filename);
 
-    DebugEvent("KFW_Logon_Event - ccache %s", cachename);
-
     DeleteFile(filename);
 
     code = pkrb5_init_context(&ctx);
@@ -944,9 +806,6 @@ KFW_copy_cache_to_system_file(char * user, char * szLogonId)
     code = pkrb5_cc_initialize(ctx, ncc, princ);
     if (code) goto cleanup;
 
-    code = KFW_set_ccache_dacl(filename, NULL);
-    if (code) goto cleanup;
-
     code = pkrb5_cc_copy_creds(ctx,cc,ncc);
 
   cleanup:
@@ -968,9 +827,9 @@ KFW_copy_cache_to_system_file(char * user, char * szLogonId)
 }
 
 int
-KFW_copy_file_cache_to_default_cache(char * filename)
+KFW_copy_system_file_to_default_cache(char * filename)
 {
-    char cachename[MAX_PATH + 8] = "FILE:";
+    char cachename[264] = "FILE:";
     krb5_context		ctx = 0;
     krb5_error_code		code;
     krb5_principal              princ = 0;
@@ -978,10 +837,10 @@ KFW_copy_file_cache_to_default_cache(char * filename)
     krb5_ccache                 ncc = 0;
     int retval = 1;
 
-    if (!pkrb5_init_context || !filename)
+    if (!pkrb5_init_context)
         return 1;
 
-    if ( strlen(filename) + sizeof("FILE:") > sizeof(cachename) )
+    if ( strlen(filename) + 6 > sizeof(cachename) )
         return 1;
 
     strcat(cachename, filename);
@@ -990,31 +849,17 @@ KFW_copy_file_cache_to_default_cache(char * filename)
     if (code) ctx = 0;
 
     code = pkrb5_cc_resolve(ctx, cachename, &cc);
-    if (code) {
-	DebugEvent0("kfwcpcc krb5_cc_resolve failed");
-	goto cleanup;
-    }
+    if (code) goto cleanup;
     
     code = pkrb5_cc_get_principal(ctx, cc, &princ);
-    if (code) {
-	DebugEvent0("kfwcpcc krb5_cc_get_principal failed");
-	goto cleanup;
-    }
+    if (code) goto cleanup;
 
     code = pkrb5_cc_default(ctx, &ncc);
-    if (code) {
-	DebugEvent0("kfwcpcc krb5_cc_default failed");
-	goto cleanup;
-    }
     if (!code) {
         code = pkrb5_cc_initialize(ctx, ncc, princ);
 
         if (!code)
             code = pkrb5_cc_copy_creds(ctx,cc,ncc);
-	if (code) {
-	    DebugEvent0("kfwcpcc krb5_cc_copy_creds failed");
-	    goto cleanup;
-	}
     }
     if ( ncc ) {
         pkrb5_cc_close(ctx, ncc);
